@@ -4,7 +4,7 @@
  * ----------------------------------------------------------------------------
  * Mastering programed by YAMASHINA Hio
  * ----------------------------------------------------------------------------
- * $Id: conv.c,v 1.7 2005/05/15 08:34:42 hio Exp $
+ * $Id: conv.c,v 1.9 2005/08/15 08:58:42 hio Exp $
  * ------------------------------------------------------------------------- */
 
 #ifdef _MSC_VER
@@ -74,21 +74,26 @@ xs_sjis_utf8(SV* sv_str)
     const unsigned char* ptr;
     if( src[0]<0x80 )
     { /* ASCII */
-      /*fprintf(stderr,"ascii: %02x\n",src[0]); */
+      ECHO_U2S((stderr,"ascii: %02x\n",src[0]));
       SV_Buf_append_ch(&result,*src);
       ++src;
       continue;
     }else if( 0xa1<=src[0] && src[0]<=0xdf )
     { /* half-width katakana (ja:Ⱦѥ) */
-      /* fprintf(stderr,"kana: %02x\n",src[0]); */
-      ptr = (unsigned char*)&g_s2u_table[src[0]];
+      ECHO_U2S((stderr,"kana: %02x\n",src[0]));
+      ptr = (unsigned char*)&g_s2u_table[(src[0]-0xa1)*3];
       ++src;
-    }else if( ((0x81<=src[0] && src[0]<=0x9f) || (0xe0<=src[0] && src[0]<=0xfc) )
-              && (0x40<=src[1] && src[1]<=0xfc && src[1]!=0x7f) )
+    }else if( src+1<src_end && 0x81<=src[0] && src[0]<=0x9f )
     { /* a double-byte letter (ja:2Хʸ) */
-      register const unsigned short sjis = ntohs(*(unsigned short*)src);
-      /*fprintf(stderr,"sjis: %04x\n",sjis); */
-      ptr = (unsigned char*)&g_s2u_table[sjis];
+      const unsigned short sjis = ntohs(*(unsigned short*)src);
+      ECHO_U2S((stderr,"sjis.dbcs#1: %04x\n",sjis));
+      ptr = (unsigned char*)&g_s2u_table[(sjis - 0x8100 + 0x3f)*3];
+      src += 2;
+    }else if( src+1<src_end && 0xe0<=src[0] && src[0]<=0xfc )
+    { /* a double-byte letter (ja:2Хʸ) */
+      const unsigned short sjis = ntohs(*(unsigned short*)src);
+      ECHO_U2S((stderr,"sjis.dbcs#2: %04x\n",sjis));
+      ptr = (unsigned char*)&g_s2u_table[(sjis- 0xe000 + 0x1f3f)*3];
       src += 2;
     }else
     { /* unknown */
@@ -98,12 +103,9 @@ xs_sjis_utf8(SV* sv_str)
       continue;
     }
 
-    /*fprintf(stderr,"utf8-char : %02x %02x %02x %02x\n",ptr[0],ptr[1],ptr[2],ptr[3]); */
-    if( ptr[3] )
-    {
-      /*fprintf(stderr,"utf8-len: [%d]\n",4); */
-      SV_Buf_append_ch4(&result,*(int*)ptr);
-    }else if( ptr[2] )
+    ECHO_U2S((stderr,"offset: 0x%04x\n",ptr-g_s2u_table));
+    ECHO_U2S((stderr,"utf8-char : %02x %02x %02x\n",ptr[0],ptr[1],ptr[2]));
+    if( ptr[2] )
     {
       /*fprintf(stderr,"utf8-len: [%d]\n",3); */
       SV_Buf_append_ch3(&result,*(int*)ptr);
@@ -156,7 +158,8 @@ xs_utf8_sjis(SV* sv_str)
 
   while( src<src_end )
   {
-    unsigned int ucs;
+    UJ_UINT32 ucs;
+    const UJ_UINT8* sjis_ptr;
     
     if( *src<=0x7f )
     {
@@ -173,10 +176,11 @@ xs_utf8_sjis(SV* sv_str)
     
     /* non-ascii */
     if( 0xe0<=*src && *src<=0xef )
-    {
+    { /* 3byte range. mostly enter here. */
       const int          utf8_len = 3;
       const unsigned int ucs_min  = 0x800;
       const unsigned int ucs_max  = 0xffff;
+      ECHO_U2S((stderr,"utf8-len: [%d]\n",utf8_len));
       /* check length */
       if( src+utf8_len<=src_end )
       { /* noop */
@@ -212,6 +216,7 @@ xs_utf8_sjis(SV* sv_str)
       const int          utf8_len = 4;
       const unsigned int ucs_min  = 0x010000;
       const unsigned int ucs_max  = 0x10ffff;
+      ECHO_U2S((stderr,"utf8-len: [%d]\n",utf8_len));
       /* check length */
       if( src+utf8_len<=src_end )
       { /* noop */
@@ -250,14 +255,19 @@ xs_utf8_sjis(SV* sv_str)
         continue;
       }
       
-      /* > U+FFFF not supported. */
-      SV_Buf_append_ch(&result,'?');
-      continue;
+      /* > U+10FFFF not supported by UTF-8 (RFC 3629). */
+      if( ucs>0x10FFFF )
+      {
+        SV_Buf_append_ch(&result,'?');
+        continue;
+      }
     }else if( 0xc0<=*src && *src<=0xdf )
     {
       const int          utf8_len = 2;
       const unsigned int ucs_min  =  0x80;
       const unsigned int ucs_max  = 0x7ff;
+      ECHO_U2S((stderr,"utf8-len: [%d]\n",utf8_len));
+      /* check length */
       if( src+utf8_len<=src_end )
       { /* noop */
       }else
@@ -291,6 +301,8 @@ xs_utf8_sjis(SV* sv_str)
     }else if( 0xf8<=*src && *src<=0xfb )
     {
       const int          utf8_len = 5;
+      ECHO_U2S((stderr,"utf8-len: [%d]\n",utf8_len));
+      /* check length */
       if( src+utf8_len<=src_end )
       { /* noop */
       }else
@@ -311,12 +323,15 @@ xs_utf8_sjis(SV* sv_str)
       }
       
       /* compute code point */
+      /* > U+10FFFF not supported by UTF-8 (RFC 3629). */
       src += utf8_len;
       SV_Buf_append_ch(&result,'?');
       continue;
     }else if( 0xfc<=*src && *src<=0xfd )
     {
       const int          utf8_len = 6;
+      ECHO_U2S((stderr,"utf8-len: [%d]\n",utf8_len));
+      /* check length */
       if( src+utf8_len<=src_end )
       { /* noop */
       }else
@@ -338,6 +353,7 @@ xs_utf8_sjis(SV* sv_str)
       }
       
       /* compute code point */
+      /* > U+10FFFF not supported by UTF-8 (RFC 3629). */
       src += utf8_len;
       SV_Buf_append_ch(&result,'?');
       continue;
@@ -349,17 +365,28 @@ xs_utf8_sjis(SV* sv_str)
     }
     
     /* ucs => sjis */
-    ECHO_U2S((stderr,"ucs2 [%04x]\n",ucs));
-    ECHO_U2S((stderr,"sjis [%04x]\n",ntohs(g_u2s_table[ucs]) ));
-    
-    if( g_u2s_table[ucs] || !ucs )
-    { /* бʸucs=='\0'λ */
-      if( g_u2s_table[ucs] & 0xff00 )
+    ECHO_U2S((stderr,"ucs [%04x]\n",ucs));
+    if( ucs<=0x9FFF ) 
+    {
+      sjis_ptr = g_u2s_table + ucs*2;
+    }else if( 0xF900<=ucs && ucs<=0xFFFF )
+    {
+      sjis_ptr = g_u2s_table + (ucs - 0xF900 + 0xA000)*2;
+    }else if( 0x0FE000<=ucs && ucs<=0x0FFFFF )
+    {
+      sjis_ptr = "?"; /* exactly 2byte: "?\0" */
+    }else
+    {
+      sjis_ptr = "\0"; /* exactly 2byte: "\0\0" */
+    }
+    if( *(const UJ_UINT16*)sjis_ptr != 0 )
+    { /* mapping dest exists. */
+      if( sjis_ptr[1]!=0 )
       {
-        SV_Buf_append_ch2(&result,g_u2s_table[ucs]);
+        SV_Buf_append_ch2(&result,*(UJ_UINT16*)sjis_ptr);
       }else
       {
-        SV_Buf_append_ch(&result,(unsigned char)g_u2s_table[ucs]);
+        SV_Buf_append_ch(&result,sjis_ptr[0]);
       }
     }else if( ucs<=0x7F )
     {
@@ -368,9 +395,9 @@ xs_utf8_sjis(SV* sv_str)
     {
       SV_Buf_append_entityref(&result,ucs);
     }
-  } /* for */
+  } /* while */
 
-  ON_U2S( bin_dump("out",result.getBegin(),result.getLength()) );
+  ON_U2S( bin_dump("out",SV_Buf_getBegin(&result),SV_Buf_getLength(&result)) );
   SV_Buf_setLength(&result);
 
   return SV_Buf_getSv(&result);
